#pragma once

#include "Uniq_ptr.h"

#include <memory>
#include <vector>

template <class T>
class Tracker
{
public:
    class Guard;

    using tag_t = T;
    using log_t = std::vector<tag_t>;
    using ptr_t = Uniq_ptr<Guard>;

    // This next line has no effect on the meaning of the code, but it
    // helps prevent us from searching for the/a constructor.
    Tracker() = default;

    log_t const& log() const;

    ptr_t operator()(const tag_t&);

private:
    using log_ptr_t = std::shared_ptr<log_t>;
    log_ptr_t log_{new log_t};
};

template <class T>
class Tracker<T>::Guard
{
public:
    Guard(Guard const&) = delete;
    Guard& operator=(Guard const&) = delete;

    ~Guard();

private:
    friend Tracker;

    Guard(log_ptr_t const&, tag_t const&);

    log_ptr_t log_;
    tag_t     tag_;
};

template <class T>
auto Tracker<T>::log() const -> log_t const&
{
    return *log_;
}

template <class T>
auto Tracker<T>::operator()(tag_t const& tag) -> ptr_t
{
    return ptr_t(new Guard(log_, tag));
}

template <class T>
Tracker<T>::Guard::Guard(log_ptr_t const& log, tag_t const& tag)
        : log_(log)
        , tag_(tag)
{ }

template <class T>
Tracker<T>::Guard::~Guard()
{
    log_->push_back(tag_);
}


